"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.HtmlBuilder = void 0;

var _fs = _interopRequireDefault(require("fs"));

var _path = _interopRequireDefault(require("path"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
class HtmlBuilder {
  constructor(rawReports, outputDir) {
    this._reportFolder = void 0;
    this._tests = new Map();
    this._reportFolder = _path.default.resolve(process.cwd(), outputDir);

    const dataFolder = _path.default.join(this._reportFolder, 'data');

    _fs.default.mkdirSync(dataFolder, {
      recursive: true
    });

    const appFolder = _path.default.join(__dirname, '..', '..', 'web', 'htmlReport2');

    for (const file of _fs.default.readdirSync(appFolder)) _fs.default.copyFileSync(_path.default.join(appFolder, file), _path.default.join(this._reportFolder, file));

    const projects = rawReports.map(rawReport => {
      const json = JSON.parse(_fs.default.readFileSync(rawReport, 'utf-8'));
      const suits = json.suites.map(s => this._createSuiteTreeItem(s));
      return {
        name: json.project.name,
        suits,
        failedTests: suits.reduce((a, s) => a + s.failedTests, 0)
      };
    });

    _fs.default.writeFileSync(_path.default.join(dataFolder, 'projects.json'), JSON.stringify(projects, undefined, 2));

    for (const [testId, test] of this._tests) {
      const testCase = {
        testId: test.testId,
        title: test.title,
        location: test.location,
        results: test.results.map(r => this._createTestResult(r))
      };

      _fs.default.writeFileSync(_path.default.join(dataFolder, testId + '.json'), JSON.stringify(testCase, undefined, 2));
    }
  }

  _createSuiteTreeItem(suite) {
    const suites = suite.suites.map(s => this._createSuiteTreeItem(s));
    const tests = suite.tests.map(t => this._createTestTreeItem(t));
    return {
      title: suite.title,
      location: suite.location,
      duration: suites.reduce((a, s) => a + s.duration, 0) + tests.reduce((a, t) => a + t.duration, 0),
      failedTests: suites.reduce((a, s) => a + s.failedTests, 0) + tests.reduce((a, t) => t.outcome === 'unexpected' || t.outcome === 'flaky' ? a + 1 : a, 0),
      suites,
      tests
    };
  }

  _createTestTreeItem(test) {
    const duration = test.results.reduce((a, r) => a + r.duration, 0);

    this._tests.set(test.testId, test);

    return {
      testId: test.testId,
      location: test.location,
      title: test.title,
      duration,
      outcome: test.outcome
    };
  }

  _createTestResult(result) {
    return {
      duration: result.duration,
      startTime: result.startTime,
      retry: result.retry,
      steps: result.steps.map(s => this._createTestStep(s)),
      error: result.error,
      status: result.status
    };
  }

  _createTestStep(step) {
    return {
      title: step.title,
      startTime: step.startTime,
      duration: step.duration,
      steps: step.steps.map(s => this._createTestStep(s)),
      log: step.log,
      error: step.error
    };
  }

}

exports.HtmlBuilder = HtmlBuilder;