"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _fs = _interopRequireDefault(require("fs"));

var _path = _interopRequireDefault(require("path"));

var _utils = require("../../utils/utils");

var _util = require("../util");

var _json = require("./json");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
class RawReporter {
  constructor() {
    this.config = void 0;
    this.suite = void 0;
  }

  onBegin(config, suite) {
    this.config = config;
    this.suite = suite;
  }

  async onEnd() {
    const projectSuites = this.suite.suites;

    for (const suite of projectSuites) {
      const project = suite._projectConfig;
      (0, _utils.assert)(project, 'Internal Error: Invalid project structure');

      const reportFolder = _path.default.join(project.outputDir, 'report');

      _fs.default.mkdirSync(reportFolder, {
        recursive: true
      });

      let reportFile;

      for (let i = 0; i < 10; ++i) {
        reportFile = _path.default.join(reportFolder, (0, _util.sanitizeForFilePath)(project.name || 'project') + (i ? '-' + i : '') + '.report');

        try {
          if (_fs.default.existsSync(reportFile)) continue;
        } catch (e) {}

        break;
      }

      if (!reportFile) throw new Error('Internal error, could not create report file');
      const report = {
        config: this.config,
        project: {
          metadata: project.metadata,
          name: project.name,
          outputDir: (0, _json.toPosixPath)(project.outputDir),
          repeatEach: project.repeatEach,
          retries: project.retries,
          testDir: (0, _json.toPosixPath)(project.testDir),
          testIgnore: (0, _json.serializePatterns)(project.testIgnore),
          testMatch: (0, _json.serializePatterns)(project.testMatch),
          timeout: project.timeout
        },
        suites: suite.suites.map(s => this._serializeSuite(s, reportFolder))
      };

      _fs.default.writeFileSync(reportFile, JSON.stringify(report, undefined, 2));
    }
  }

  _serializeSuite(suite, reportFolder) {
    return {
      title: suite.title,
      location: suite.location,
      suites: suite.suites.map(s => this._serializeSuite(s, reportFolder)),
      tests: suite.tests.map(t => this._serializeTest(t, reportFolder))
    };
  }

  _serializeTest(test, reportFolder) {
    const testId = (0, _utils.calculateSha1)(test.titlePath().join('|'));
    return {
      testId,
      title: test.title,
      location: test.location,
      expectedStatus: test.expectedStatus,
      timeout: test.timeout,
      annotations: test.annotations,
      retries: test.retries,
      ok: test.ok(),
      outcome: test.outcome(),
      results: test.results.map(r => this._serializeResult(testId, test, r, reportFolder))
    };
  }

  _serializeResult(testId, test, result, reportFolder) {
    return {
      retry: result.retry,
      workerIndex: result.workerIndex,
      startTime: result.startTime.toISOString(),
      duration: result.duration,
      status: result.status,
      error: result.error,
      attachments: this._createAttachments(reportFolder, testId, result),
      steps: this._serializeSteps(test, result.steps)
    };
  }

  _serializeSteps(test, steps) {
    return steps.map(step => {
      return {
        title: step.title,
        category: step.category,
        startTime: step.startTime.toISOString(),
        duration: step.duration,
        error: step.error,
        steps: this._serializeSteps(test, step.steps),
        log: step.data.log || undefined
      };
    });
  }

  _createAttachments(reportFolder, testId, result) {
    const attachments = [];

    for (const attachment of result.attachments.filter(a => !a.path)) {
      const sha1 = (0, _utils.calculateSha1)(attachment.body);

      const file = _path.default.join(reportFolder, sha1);

      try {
        _fs.default.writeFileSync(_path.default.join(reportFolder, sha1), attachment.body);

        attachments.push({
          name: attachment.name,
          contentType: attachment.contentType,
          path: (0, _json.toPosixPath)(file)
        });
      } catch (e) {}
    }

    for (const attachment of result.attachments.filter(a => a.path)) attachments.push(attachment);

    if (result.stdout.length) attachments.push(this._stdioAttachment(reportFolder, testId, result, 'stdout'));
    if (result.stderr.length) attachments.push(this._stdioAttachment(reportFolder, testId, result, 'stderr'));
    return attachments;
  }

  _stdioAttachment(reportFolder, testId, result, type) {
    const file = `${testId}.${result.retry}.${type}`;

    const fileName = _path.default.join(reportFolder, file);

    for (const chunk of type === 'stdout' ? result.stdout : result.stderr) {
      if (typeof chunk === 'string') _fs.default.appendFileSync(fileName, chunk + '\n');else _fs.default.appendFileSync(fileName, chunk);
    }

    return {
      name: type,
      contentType: 'application/octet-stream',
      path: (0, _json.toPosixPath)(fileName)
    };
  }

}

var _default = RawReporter;
exports.default = _default;